syntax = "proto3";
package grpctest;

message TestRequest {
    int32 Input = 1;
}

message TestResponse {
    int32 Output = 2;
}

service Test {
    rpc Double(TestRequest) returns (TestResponse) {}
}
